import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcfractal.testing_helpers import mname1
from qcportal import PortalRequestError
from qcportal.auth import RoleInfo, UserInfo
from qcportal.molecules import Molecule


@pytest.fixture(scope="module")
def module_authtest_snowflake(postgres_server, pytestconfig):
    """
    A QCFractal snowflake server for testing authorization
    """

    pg_harness = postgres_server.get_new_harness("module_authtest_snowflake")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(
        pg_harness,
        encoding,
        start_api=True,
        create_users=False,
        enable_security=True,
        allow_unauthenticated_read=False,
    ) as snowflake:
        pg_harness.create_template()
        yield snowflake


@pytest.fixture(scope="function")
def authtest_snowflake(module_authtest_snowflake):
    """
    A QCFractal snowflake used for testing (function-scoped)

    The underlying database is deleted and recreated after each test function
    """

    yield module_authtest_snowflake
    module_authtest_snowflake.reset()


def test_auth_allow_deny_read(authtest_snowflake):
    # add a test role. Can only access molecule
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket = authtest_snowflake.get_storage_socket()
    socket.roles.add(rinfo)

    uinfo = UserInfo(username="test_user", enabled=True, role="test_role")
    pw = socket.users.add(uinfo)

    client = authtest_snowflake.client(username="test_user", password=pw)
    client.query_molecules()

    ####################################################
    # Explicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Deny", "Action": "READ", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.query_molecules()

    ####################################################
    # Implicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.query_molecules()


def test_auth_allow_deny_write(authtest_snowflake):
    test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])

    # add a test role. Can only access molecule
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Allow", "Action": "WRITE", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket = authtest_snowflake.get_storage_socket()
    socket.roles.add(rinfo)

    uinfo = UserInfo(username="test_user", enabled=True, role="test_role")
    pw = socket.users.add(uinfo)

    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.query_molecules()

    client.add_molecules([test_mol])

    ####################################################
    # Explicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Deny", "Action": "WRITE", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.add_molecules([test_mol])

    ####################################################
    # Implicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.add_molecules([test_mol])


def test_auth_allow_deny_delete(authtest_snowflake):
    # add a test role. Can only access molecule
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Allow", "Action": "DELETE", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket = authtest_snowflake.get_storage_socket()
    socket.roles.add(rinfo)

    uinfo = UserInfo(username="test_user", enabled=True, role="test_role")
    pw = socket.users.add(uinfo)

    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.query_molecules()

    client.delete_molecules([1])

    ####################################################
    # Explicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
                {"Effect": "Deny", "Action": "DELETE", "Resource": ["/api/v1/molecules"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.delete_molecules([1])

    ####################################################
    # Implicit deny
    rinfo = RoleInfo(
        rolename="test_role",
        permissions={
            "Statement": [
                {"Effect": "Allow", "Action": "READ", "Resource": ["/api/v1/information"]},
            ]
        },
    )

    socket.roles.modify(rinfo)
    client = authtest_snowflake.client(username="test_user", password=pw)

    with pytest.raises(PortalRequestError, match=r"not authorized to access"):
        client.delete_molecules([1])


def test_auth_default_role_read(secure_snowflake):
    client = secure_snowflake.client("read_user", password=test_users["read_user"]["pw"])

    test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])

    # Read from molecules and records
    client.query_molecules()
    client.query_records()

    # Can't add molecules or records
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_molecules([test_mol])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_singlepoints(test_mol, "prog", "energy", "b3lyp", "6-31g")

    # Or modify things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_molecule(0, name="new name")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_records([0], new_tag="new tag")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.reset_records([0])

    # Or delete things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_molecules([0])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_records([0])

    # Can't access users and roles
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_users()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_user("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_roles()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_role("read")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user", "new_pw")

    # Can't access error log, server log, etc
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_access_log()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_error_log()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_managers()

    # Can't activate a manager
    mclient = secure_snowflake.manager_client(mname1, "read_user", test_users["read_user"]["pw"])
    with pytest.raises(PortalRequestError, match="Forbidden"):
        mclient.activate("1.23", {"prog": ["1.23"]}, ["*"])

    # Can access own user, and modify it
    uinfo = client.get_user()
    uinfo.fullname = "A new full name"
    client.modify_user(uinfo)
    client.change_user_password()


def test_auth_default_role_submit(secure_snowflake):
    client = secure_snowflake.client("submit_user", password=test_users["submit_user"]["pw"])

    test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])

    # Read from molecules and records
    client.query_molecules()
    client.query_records()

    # Can add molecules or records
    _, mol_ids = client.add_molecules([test_mol])
    _, rec_ids = client.add_singlepoints(test_mol, "prog", "energy", "b3lyp", "6-31g")

    # and modify things
    client.modify_molecule(mol_ids[0], name="new name")
    client.modify_records(rec_ids[0], new_tag="new tag")
    client.reset_records(rec_ids)

    # and delete things
    client.delete_molecules(mol_ids)
    client.delete_records(rec_ids)

    # Can't access users and roles
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_users()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_user("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_roles()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_role("read")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user", "new_pw")

    # Can't access error log, server log, etc
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_access_log()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_error_log()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_managers()

    # Can't activate a manager
    mclient = secure_snowflake.manager_client(mname1, "submit_user", test_users["submit_user"]["pw"])
    with pytest.raises(PortalRequestError, match="Forbidden"):
        mclient.activate("1.23", {"prog": ["1.23"]}, ["*"])

    # Can access own user, and modify it
    uinfo = client.get_user()
    uinfo.fullname = "A new full name"
    client.modify_user(uinfo)
    client.change_user_password()


def test_auth_default_role_monitor(secure_snowflake):
    client = secure_snowflake.client("monitor_user", password=test_users["monitor_user"]["pw"])

    test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])

    # Read from molecules and records
    client.query_molecules()
    client.query_records()

    # Can't add molecules or records
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_molecules([test_mol])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_singlepoints(test_mol, "prog", "energy", "b3lyp", "6-31g")

    # Or modify things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_molecule(0, name="new name")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_records([0], new_tag="new tag")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.reset_records([0])

    # Or delete things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_molecules([0])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_records([0])

    # Can't access users and roles
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_users()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_user("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_roles()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_role("read")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user", "new_pw")

    # Can access error log, server log, etc
    client.query_access_log()
    client.query_error_log()
    client.query_managers()

    # Can't activate a manager
    mclient = secure_snowflake.manager_client(mname1, "monitor_user", test_users["monitor_user"]["pw"])
    with pytest.raises(PortalRequestError, match="Forbidden"):
        mclient.activate("1.23", {"prog": ["1.23"]}, ["*"])

    # Can access own user, and modify it
    uinfo = client.get_user()
    uinfo.fullname = "A new full name"
    client.modify_user(uinfo)
    client.change_user_password()


def test_auth_default_role_compute(secure_snowflake):
    client = secure_snowflake.client("compute_user", password=test_users["compute_user"]["pw"])

    test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])

    # Can't read from molecules and records
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_molecules()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_records()

    # Can't add molecules or records
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_molecules([test_mol])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.add_singlepoints(test_mol, "prog", "energy", "b3lyp", "6-31g")

    # Or modify things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_molecule(0, name="new name")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.modify_records([0], new_tag="new tag")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.reset_records([0])

    # Or delete things
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_molecules([0])

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.delete_records([0])

    # Can't access users and roles
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_users()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_user("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.list_roles()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.get_role("read")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user")

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.change_user_password("admin_user", "new_pw")

    # Can't access error log, server log, etc
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_access_log()

    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_error_log()

    # Can't query managers
    with pytest.raises(PortalRequestError, match="Forbidden"):
        client.query_managers()

    # Can activate a manager
    mclient = secure_snowflake.manager_client(mname1, "compute_user", test_users["compute_user"]["pw"])
    mclient.activate("1.23", {"prog": ["1.23"]}, ["*"])

    # Can access own user, and modify it
    uinfo = client.get_user()
    uinfo.fullname = "A new full name"
    client.modify_user(uinfo)
    client.change_user_password()



"""
    Credit: https://github.com/Netflix-Skunkworks/policyuniverse
    Patrick Kelley <patrick@netflix.com>

"""

from .policy import Policy


policy01 = dict(
    Version="2012-10-08",
    Statement=dict(
        Effect="Allow",
        Principal="*",
        Action=["rds:*"],
        Resource="*",
    ),
)

policy02 = dict(
    Version="2010-08-14",
    Statement=[
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::012345678910:root",
            Action=["rds:*"],
            Resource="*",
        )
    ],
)

# One statement limits by ARN, the other allows any account number
policy03 = dict(
    Version="2010-08-14",
    Statement=[
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::012345678910:root",
            Action=["s3:*"],
            Resource="*",
        ),
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::*:role/Hello",
            Action=["ec2:*"],
            Resource="*",
        ),
    ],
)

policy04 = dict(
    Version="2010-08-14",
    Statement=[
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::012345678910:root",
            Action=["s3:*"],
            Resource="*",
        ),
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::*:role/Hello",
            Action=["ec2:*"],
            Resource="*",
        ),
    ],
)

policy05 = dict(
    Version="2010-08-14",
    Statement=[
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::012345678910:root",
            Action=["s3:*"],
            Resource="*",
        ),
        dict(
            Effect="Allow",
            Principal="arn:aws:iam::*:role/Hello",
            Action=["ec2:*"],
            Resource="*",
        ),
    ],
)

# AWS Organizations
policy06 = dict(
    Version="2010-08-14",
    Statement=[
        dict(
            Effect="Allow",
            Principal="*",
            Action=["rds:*"],
            Resource="*",
        )
    ],
)


def test_principals():
    assert Policy(policy04).principals == set(["arn:aws:iam::012345678910:root", "arn:aws:iam::*:role/Hello"])


def test_whos_allowed():
    allowed = Policy(policy03).whos_allowed()
    assert len(allowed) == 2

    allowed = Policy(policy04).whos_allowed()
    assert len(allowed) == 2
    principal_allowed = set([item for item in allowed if item.category == "principal"])
    assert len(principal_allowed) == 2

    allowed = Policy(policy06).whos_allowed()
    assert len(allowed) == 1



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Tuple, Optional, Dict, Union, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcelemental.models import Molecule, FailedOperation, ComputeError, OptimizationResult
from qcelemental.models.procedures import OptimizationProtocols

from qcarchivetesting.helpers import read_record_data
from qcfractal.components.gridoptimization.record_db_models import GridoptimizationRecordORM
from qcfractal.testing_helpers import run_service
from qcportal.gridoptimization import GridoptimizationSpecification, GridoptimizationKeywords
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordTask
from qcportal.singlepoint import SinglepointProtocols, QCSpecification
from qcportal.utils import recursive_normalizer

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName


def compare_gridoptimization_specs(
    input_spec: Union[GridoptimizationSpecification, Dict[str, Any]],
    output_spec: Union[GridoptimizationSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = GridoptimizationSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = GridoptimizationSpecification(**output_spec)

    return input_spec == output_spec


test_specs = [
    GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
            ],
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    ),
    GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=True,
            scans=[
                {"type": "dihedral", "indices": [3, 2, 1, 0], "steps": [-90, -45, 0, 45, 90], "step_type": "absolute"},
            ],
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all", stdout=False),
            ),
        ),
    ),
]


def generate_task_key(task: RecordTask):
    # task is an optimization
    inp_data = task.function_kwargs["input_data"]
    assert inp_data["schema_name"] in "qcschema_optimization_input"

    mol_hash = inp_data["initial_molecule"]["identifiers"]["molecule_hash"]
    constraints = inp_data["keywords"].get("constraints", None)

    # Lookups may depend on floating point values
    constraints = recursive_normalizer(constraints)

    # This is the key in the dictionary of optimization results
    constraints_str = json.dumps(constraints, sort_keys=True)
    return mol_hash + "|" + constraints_str


def load_test_data(name: str) -> Tuple[GridoptimizationSpecification, Molecule, Dict[str, OptimizationResult]]:
    test_data = read_record_data(name)

    return (
        pydantic.parse_obj_as(GridoptimizationSpecification, test_data["specification"]),
        pydantic.parse_obj_as(Molecule, test_data["initial_molecule"]),
        pydantic.parse_obj_as(Dict[str, OptimizationResult], test_data["results"]),
    )


def submit_test_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    tag: Optional[str] = "*",
    priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, Dict[str, OptimizationResult]]:
    input_spec, molecule, result = load_test_data(name)
    meta, record_ids = storage_socket.records.gridoptimization.add(
        [molecule], input_spec, tag, priority, None, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_test_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    tag: Optional[str] = "*",
    priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    record_id, result = submit_test_data(storage_socket, name, tag, priority)

    with storage_socket.session_scope() as session:
        record = session.get(GridoptimizationRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        result = {x: failed_op for x in result}

    finished, n_optimizations = run_service(storage_socket, manager_name, record_id, generate_task_key, result, 200)
    assert finished

    with storage_socket.session_scope() as session:
        record = session.get(GridoptimizationRecordORM, record_id)
        assert record.status == end_status

    return record_id



from __future__ import annotations

import itertools
import logging
import math
from typing import List, Dict, Tuple, Optional, Sequence, Any, Union, Set, TYPE_CHECKING

import tabulate
from sqlalchemy import select
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import defer, undefer, lazyload, joinedload

from qcfractal import __version__ as qcfractal_version
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.manybody import (
    BSSECorrectionEnum,
    ManybodyKeywords,
    ManybodySpecification,
    ManybodyQueryFilters,
)
from qcportal.metadata_models import InsertMetadata
from qcportal.molecules import Molecule
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import hash_dict
from .record_db_models import ManybodyClusterORM, ManybodyRecordORM, ManybodySpecificationORM
from ..record_socket import BaseRecordSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket


# Meaningless, but unique to manybody
manybody_insert_lock_id = 14500


def nCr(n: int, r: int) -> int:
    """
    Compute the binomial coefficient n! / (k! * (n-k)!)
    """

    # TODO: available in python 3.8 as math.comb
    return math.factorial(n) // (math.factorial(r) * math.factorial(n - r))


def analyze_results(mb_orm: ManybodyRecordORM):
    keywords = ManybodyKeywords(**mb_orm.specification.keywords)

    # Total number of fragments present on the molecule
    total_frag = len(mb_orm.initial_molecule.fragments)

    # Group clusters by nbody
    # For CP, this only includes the calculations done in the full basis
    clusters = {}
    for c in mb_orm.clusters:
        if keywords.bsse_correction == BSSECorrectionEnum.none:
            nbody = len(c.fragments)
            clusters.setdefault(nbody, [])
            clusters[nbody].append(c)
        elif keywords.bsse_correction == BSSECorrectionEnum.cp and len(c.basis) > 1:
            nbody = len(c.fragments)
            clusters.setdefault(nbody, [])
            clusters[nbody].append(c)

    # Total energy for each nbody cluster. This is the energy calculated
    # by the singlepoint multiplied by its degeneracy
    cluster_energy: Dict[int, float] = {}

    for nbody, v in clusters.items():
        cluster_energy[nbody] = sum(c.degeneracy * c.singlepoint_record.properties["return_energy"] for c in v)

    # Calculate CP correction
    bsse = 0.0
    if keywords.bsse_correction == BSSECorrectionEnum.cp:
        monomer_clusters = [c for c in mb_orm.clusters if len(c.fragments) == 1 and len(c.basis) == 1]
        monomer_energy = sum(c.degeneracy * c.singlepoint_record.properties["return_energy"] for c in monomer_clusters)
        bsse = cluster_energy[1] - monomer_energy

    # Total energies
    total_energy_through = {}

    for n in cluster_energy.keys():
        # If entire molecule was calculated, then add that
        if n == total_frag:
            total_energy_through[n] = cluster_energy[n]
        elif n == 1:
            total_energy_through[n] = cluster_energy[n]
        else:
            total_energy_through[n] = 0.0
            for nbody in range(1, n + 1):
                sign = (-1) ** (n - nbody)
                take_nk = nCr(total_frag - nbody - 1, n - nbody)
                total_energy_through[n] += take_nk * sign * cluster_energy[nbody]

    # Apply CP correction
    if keywords.bsse_correction == BSSECorrectionEnum.cp:
        total_energy_through = {k: v - bsse for k, v in total_energy_through.items()}

    # Contributions to interaction energy
    energy_contrib = {}
    energy_contrib[1] = 0.0
    for n in total_energy_through:
        if n != 1:
            energy_contrib[n] = total_energy_through[n] - total_energy_through[n - 1]

    # Interaction energy
    interaction_energy = {}
    for n in total_energy_through:
        interaction_energy[n] = total_energy_through[n] - total_energy_through[1]

    results = {
        "cluster_energy": cluster_energy,
        "total_energy_through": total_energy_through,
        "interaction_energy": interaction_energy,
        "energy_contrib": energy_contrib,
    }

    mb_orm.results = results


def build_mbe_clusters(mol: Molecule, keywords: ManybodyKeywords) -> List[Tuple[Set[int], Set[int], Molecule]]:
    """
    Fragments a larger molecule into clusters

    Parameters
    ----------
    mol
        Molecule to fragment
    keywords
        Keywords that control the fragmenting

    Returns
    -------
    :
        A list of tuples with three elements -
        (1) Set of fragment indices (2) Set of basis indices (3) Fragment molecule
    """

    # List: (fragments, basis, Molecule)
    # fragments and basis are sequences
    ret: List[Tuple[Set[int], Set[int], Molecule]] = []

    if len(mol.fragments) < 2:
        raise RuntimeError("manybody service: Molecule must have at least two fragments")

    max_nbody = keywords.max_nbody

    if max_nbody is None:
        max_nbody = len(mol.fragments)
    else:
        max_nbody = min(max_nbody, len(mol.fragments))

    # Build some info
    allfrag = set(range(max_nbody))

    # Loop over the nbody (the number of bodies to include. 1 = monomers, 2 = dimers)
    for nbody in range(1, max_nbody):
        for frag_idx in itertools.combinations(allfrag, nbody):
            frag_idx = set(frag_idx)
            if keywords.bsse_correction == BSSECorrectionEnum.none:
                frag_mol = mol.get_fragment(frag_idx, orient=True, group_fragments=True)
                ret.append((frag_idx, frag_idx, frag_mol))
            elif keywords.bsse_correction == BSSECorrectionEnum.cp:
                ghost = list(set(allfrag) - set(frag_idx))
                frag_mol = mol.get_fragment(frag_idx, ghost, orient=True, group_fragments=True)
                ret.append((frag_idx, allfrag, frag_mol))
            else:
                raise RuntimeError(f"Unknown BSSE correction method: {keywords.bsse_correction}")

    # Include full molecule as well
    if max_nbody >= len(mol.fragments):
        ret.append((allfrag, allfrag, mol))

    # Always include monomer in monomer basis for CP
    if keywords.bsse_correction == BSSECorrectionEnum.cp:
        for frag_idx in allfrag:
            frag_mol = mol.get_fragment([frag_idx], orient=True, group_fragments=True)
            ret.append(({frag_idx}, {frag_idx}, frag_mol))

    return ret


class ManybodyRecordSocket(BaseRecordSocket):
    """
    Socket for handling manybody computations
    """

    # Used by the base class
    record_orm = ManybodyRecordORM

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    @staticmethod
    def get_children_select() -> List[Any]:
        stmt = select(
            ManybodyClusterORM.manybody_id.label("parent_id"),
            ManybodyClusterORM.singlepoint_id.label("child_id"),
        )
        return [stmt]

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        mb_orm: ManybodyRecordORM = service_orm.record

        output = "\n\nCreated manybody calculation\n"

        output += "-" * 80 + "\nManybody Keywords:\n\n"
        spec: ManybodySpecification = mb_orm.specification.to_model(ManybodySpecification)
        table_rows = sorted(spec.keywords.dict().items())
        output += tabulate.tabulate(table_rows, headers=["keyword", "value"])
        output += "\n\n" + "-" * 80 + "\nQC Specification:\n\n"
        table_rows = sorted(spec.singlepoint_specification.dict().items())
        output += tabulate.tabulate(table_rows, headers=["keyword", "value"])
        output += "\n\n"

        init_mol: Molecule = mb_orm.initial_molecule.to_model(Molecule)

        output += f"Initial molecule: formula={init_mol.get_molecular_formula()} id={mb_orm.initial_molecule_id}\n"
        output += f"Initial molecule has {len(init_mol.fragments)} fragments\n"

        # Fragment the initial molecule into clusters
        keywords = ManybodyKeywords(**mb_orm.specification.keywords)
        mol_clusters = build_mbe_clusters(init_mol, keywords)

        output += f"Molecule is split into into {len(mol_clusters)} separate clusters:\n\n"

        # Group by nbody and count for output
        mol_clusters_nbody = {}
        for mc in mol_clusters:
            nbody = len(mc[0])
            mol_clusters_nbody.setdefault(nbody, 0)
            mol_clusters_nbody[nbody] += 1

        table_rows = [(k, v) for k, v in sorted(mol_clusters_nbody.items())]
        output += tabulate.tabulate(table_rows, headers=["n-body", "count"])
        output += "\n\n"

        # Add the manybody molecules/clusters to the db
        nbody_mols = [x[2] for x in mol_clusters]
        meta, mol_ids = self.root_socket.molecules.add(nbody_mols)

        if not meta.success:
            raise RuntimeError("Unable to add molecules to the database: " + meta.error_string)

        # We do unique ids only
        # Some manybody calculations will have identical molecules
        # Think of single-atom dimers or something. There will only be one monomer
        done_ids = set()

        table_rows = []
        for (frag_idx, basis_idx, frag_mol), mol_id in zip(mol_clusters, mol_ids):
            if mol_id in done_ids:
                continue

            done_ids.add(mol_id)
            degen = mol_ids.count(mol_id)
            frag_idx = sorted(frag_idx)
            basis_idx = sorted(basis_idx)

            new_mb_orm = ManybodyClusterORM(fragments=frag_idx, basis=basis_idx, molecule_id=mol_id, degeneracy=degen)

            mb_orm.clusters.append(new_mb_orm)

            table_rows.append((degen, frag_mol.get_molecular_formula(), mol_id, frag_idx, basis_idx))

        # Sort rows by nbody (# of fragments), the degeneracy descending, then molecule id
        table_rows = sorted(table_rows, key=lambda x: (len(x[3]), -x[0], x[2]))
        output += tabulate.tabulate(table_rows, headers=["degeneracy", "molecule", "molecule id", "fragments", "basis"])
        output += "\n\n"

        self.root_socket.records.append_output(session, mb_orm, OutputTypeEnum.stdout, output)

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ):
        mb_orm: ManybodyRecordORM = service_orm.record

        # Always update with the current provenance
        mb_orm.compute_history[-1].provenance = {
            "creator": "qcfractal",
            "version": qcfractal_version,
            "routine": "qcfractal.services.manybody",
        }

        # Grab all the clusters for the computation and them map them to molecule ID
        clusters = mb_orm.clusters
        clusters_by_mol = {c.molecule_id: c for c in clusters}

        service_orm.dependencies = []

        # What molecules/clusters we still have to do
        mols_to_compute = [c.molecule_id for c in clusters if c.singlepoint_id is None]

        if mols_to_compute:
            sp_spec_id = mb_orm.specification.singlepoint_specification_id
            meta, sp_ids = self.root_socket.records.singlepoint.add_internal(
                mols_to_compute,
                sp_spec_id,
                service_orm.tag,
                service_orm.priority,
                mb_orm.owner_user_id,
                mb_orm.owner_group_id,
                service_orm.find_existing,
                session=session,
            )

            output = f"\nSubmitted {len(sp_ids)} singlepoint calculations "
            output += f"({meta.n_inserted} new, {meta.n_existing} existing):\n\n"

            for mol_id, sp_id in zip(mols_to_compute, sp_ids):
                svc_dep = ServiceDependencyORM(record_id=sp_id, extras={})

                cluster_orm = clusters_by_mol[mol_id]

                # Assign the singlepoint id to the cluster
                assert cluster_orm.singlepoint_id is None
                cluster_orm.singlepoint_id = sp_id

                service_orm.dependencies.append(svc_dep)

            table_rows = sorted(zip(mols_to_compute, sp_ids))
            output += tabulate.tabulate(table_rows, headers=["molecule id", "singlepoint id"])

        else:
            output = "\n\n" + "*" * 80 + "\n"
            output += "All manybody singlepoint computations are complete!\n\n"

            output += "Singlepoint results:\n"

            # Map molecule_id -> singlepoint record
            result_map = {c.molecule_id: c.singlepoint_record for c in clusters}

            # Make a nice output table
            table_rows = []
            for component in mb_orm.clusters:
                mol_id = component.molecule_id
                mol_form = component.molecule.identifiers["molecular_formula"]

                energy = component.singlepoint_record.properties["return_energy"]
                table_row = [mol_id, component.singlepoint_id, mol_form, energy]
                table_rows.append(table_row)

                result_map[mol_id] = component.singlepoint_record

            output += tabulate.tabulate(
                table_rows, headers=["molecule id", "singlepoint id", "molecule", "energy (hartree)"], floatfmt=".10f"
            )

            # Create the results of the manybody calculation
            analyze_results(mb_orm)

            # Make a results table
            r = mb_orm.results
            nb_keys = sorted(r["total_energy_through"].keys())
            table_rows = [
                (
                    nbody,
                    r["total_energy_through"][nbody],
                    r["interaction_energy"][nbody],
                    r["energy_contrib"][nbody],
                )
                for nbody in nb_keys
            ]

            output += "\n\n\n\n" + "=" * 80 + "\n"
            output += "Final energy results (in hartrees)\n" + "=" * 80 + "\n\n"
            output += tabulate.tabulate(
                table_rows,
                headers=["\nnbody", "Total Energy  \nthrough n-body", "\nInteraction Energy", "\nContrib to IE"],
                floatfmt="6.10f",
            )

        self.root_socket.records.append_output(session, mb_orm, OutputTypeEnum.stdout, output)

        return len(mols_to_compute) == 0

    def add_specification(
        self, mb_spec: ManybodySpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for a manybody service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        mb_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        kw_dict = mb_spec.keywords.dict()
        kw_hash = hash_dict(kw_dict)

        with self.root_socket.optional_session(session) as session:
            meta, sp_spec_id = self.root_socket.records.singlepoint.add_specification(
                qc_spec=mb_spec.singlepoint_specification, session=session
            )

            if not meta.success:
                return (
                    InsertMetadata(
                        error_description="Unable to add singlepoint specification: " + meta.error_string,
                    ),
                    None,
                )

            stmt = (
                insert(ManybodySpecificationORM)
                .values(
                    program=mb_spec.program,
                    singlepoint_specification_id=sp_spec_id,
                    keywords=kw_dict,
                    keywords_hash=kw_hash,
                )
                .on_conflict_do_nothing()
                .returning(ManybodySpecificationORM.id)
            )

            r = session.execute(stmt).scalar_one_or_none()
            if r is not None:
                return InsertMetadata(inserted_idx=[0]), r
            else:
                # Specification was already existing
                stmt = select(ManybodySpecificationORM.id).filter_by(
                    program=mb_spec.program,
                    singlepoint_specification_id=sp_spec_id,
                    keywords_hash=kw_hash,
                )

                r = session.execute(stmt).scalar_one()
                return InsertMetadata(existing_idx=[0]), r

    def query(
        self,
        query_data: ManybodyQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Query manybody records

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of record ids that were found in the database.
        """

        and_query = []
        need_spec_join = False
        need_qcspec_join = False

        if query_data.program is not None:
            and_query.append(ManybodySpecificationORM.program.in_(query_data.program))
            need_spec_join = True
        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_program))
            need_qcspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_qcspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_qcspec_join = True
        if query_data.initial_molecule_id is not None:
            and_query.append(ManybodyRecordORM.initial_molecule_id.in_(query_data.initial_molecule_id))

        stmt = select(ManybodyRecordORM.id)

        if need_spec_join or need_qcspec_join:
            stmt = stmt.join(ManybodyRecordORM.specification)

        if need_qcspec_join:
            stmt = stmt.join(ManybodySpecificationORM.singlepoint_specification)

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=ManybodyRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_molecule_ids: Sequence[int],
        mb_spec_id: int,
        tag: str,
        priority: PriorityEnum,
        owner_user_id: Optional[int],
        owner_group_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new manybody computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecule_ids
            IDs of the initial molecules
        mb_spec_id
            ID of the specification
        tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        priority
            The priority for the computation
        owner_user_id
            ID of the user who owns the record
        owner_group_id
            ID of the group with additional permission for these records
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        tag = tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            self.root_socket.users.assert_group_member(owner_user_id, owner_group_id, session=session)

            all_orm = []

            for mid in initial_molecule_ids:
                mb_orm = ManybodyRecordORM(
                    is_service=True,
                    specification_id=mb_spec_id,
                    initial_molecule_id=mid,
                    status=RecordStatusEnum.waiting,
                    owner_user_id=owner_user_id,
                    owner_group_id=owner_group_id,
                )

                self.create_service(mb_orm, tag, priority, find_existing)
                all_orm.append(mb_orm)

            if find_existing:
                meta, ids = insert_general(
                    session,
                    all_orm,
                    (ManybodyRecordORM.specification_id, ManybodyRecordORM.initial_molecule_id),
                    (ManybodyRecordORM.id,),
                    lock_id=manybody_insert_lock_id,
                )
                return meta, [x[0] for x in ids]
            else:
                session.add_all(all_orm)
                session.flush()
                meta = InsertMetadata(inserted_idx=list(range(len(all_orm))))

                return meta, [x.id for x in all_orm]

    def add(
        self,
        initial_molecules: Sequence[Union[int, Molecule]],
        mb_spec: ManybodySpecification,
        tag: str,
        priority: PriorityEnum,
        owner_user: Optional[Union[int, str]],
        owner_group: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new manybody calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecules
            Initial molecules (objects or ids) to compute using the specification
        mb_spec
            Specification for the calculations
        tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        priority
            The priority for the computation
        owner_user
            Name or ID of the user who owns the record
        owner_group
            Group with additional permission for these records
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            owner_user_id, owner_group_id = self.root_socket.users.get_owner_ids(
                owner_user, owner_group, session=session
            )

            # First, add the specification
            spec_meta, spec_id = self.add_specification(mb_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            mol_meta, mol_ids = self.root_socket.molecules.add_mixed(initial_molecules, session=session)
            if not mol_meta.success:
                return (
                    InsertMetadata(error_description="Aborted - could not add all molecules: " + mol_meta.error_string),
                    [],
                )

            return self.add_internal(
                mol_ids, spec_id, tag, priority, owner_user_id, owner_group_id, find_existing, session=session
            )

    ####################################################
    # Some stuff to be retrieved for manybodys
    ####################################################

    def get_clusters(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(ManybodyRecordORM.clusters).options(undefer("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(ManybodyRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.clusters]



from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    TIMESTAMP,
    Float,
    Index,
    String,
    JSON,
    Enum,
    UniqueConstraint,
    CheckConstraint,
)
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import relationship

from qcfractal.db_socket import BaseORM
from qcportal.managers import ManagerStatusEnum
from qcportal.utils import now_at_utc


class ComputeManagerLogORM(BaseORM):
    """
    Table for storing manager logs

    This contains information about a manager at a particular point in time. This table
    is periodically appended to, with updated information about a manager.
    """

    __tablename__ = "compute_manager_log"

    id = Column(Integer, primary_key=True)
    manager_id = Column(Integer, ForeignKey("compute_manager.id", ondelete="cascade"), nullable=False)

    timestamp = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)

    claimed = Column(Integer, nullable=False)
    successes = Column(Integer, nullable=False)
    failures = Column(Integer, nullable=False)
    rejected = Column(Integer, nullable=False)

    active_tasks = Column(Integer, nullable=False, default=0)
    active_cores = Column(Integer, nullable=False, default=0)
    active_memory = Column(Float, nullable=False, default=0.0)
    total_cpu_hours = Column(Float, nullable=False, default=0.0)

    __table_args__ = (Index("ix_compute_manager_log_manager_id", "manager_id"),)


class ComputeManagerORM(BaseORM):
    """
    Table for storing information about active and inactive compute managers
    """

    __tablename__ = "compute_manager"

    id = Column(Integer, primary_key=True)

    name = Column(String, nullable=False)
    cluster = Column(String, nullable=False)
    hostname = Column(String, nullable=False)
    username = Column(String)  # Can be null (ie, security not enabled, snowflake)
    tags = Column(ARRAY(String), nullable=False)

    # Latest count
    claimed = Column(Integer, nullable=False, default=0)
    successes = Column(Integer, nullable=False, default=0)
    failures = Column(Integer, nullable=False, default=0)
    rejected = Column(Integer, nullable=False, default=0)

    active_tasks = Column(Integer, nullable=False, default=0)
    active_cores = Column(Integer, nullable=False, default=0)
    active_memory = Column(Float, nullable=False, default=0.0)
    total_cpu_hours = Column(Float, nullable=False, default=0.0)

    status = Column(Enum(ManagerStatusEnum), nullable=False)

    created_on = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    modified_on = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)

    manager_version = Column(String, nullable=False)
    programs = Column(JSON, nullable=False)

    log = relationship(
        ComputeManagerLogORM,
        order_by=ComputeManagerLogORM.timestamp.desc(),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __table_args__ = (
        Index("ix_compute_manager_status", "status"),
        Index("ix_compute_manager_modified_on", "modified_on", postgresql_using="brin"),
        UniqueConstraint("name", name="ux_compute_manager_name"),
        CheckConstraint("programs::text = LOWER(programs::text)", name="ck_compute_manager_programs_lower"),
        CheckConstraint("tags::text = LOWER(tags::text)", name="ck_compute_manager_tags_lower"),
    )



from qcfractal.db_socket import SQLAlchemySocket
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols
from qcportal.torsiondrive import TorsiondriveSpecification, TorsiondriveKeywords


def test_torsiondrive_socket_basic_specification(storage_socket: SQLAlchemySocket):
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 14)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog2",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver=SinglepointDriver.hessian,
                method="hf",
                basis="def2-tzvp",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    spec3 = TorsiondriveSpecification(
        # Not putting in program
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[5],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog2",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(trajectory="none"),
            qc_specification=QCSpecification(
                program="prog2",
                driver=SinglepointDriver.hessian,
                method="hf",
                basis="def2-tzvp",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
            ),
        ),
    )

    meta1, id1 = storage_socket.records.torsiondrive.add_specification(spec1)
    meta2, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    meta3, id3 = storage_socket.records.torsiondrive.add_specification(spec3)
    assert meta1.success
    assert meta2.success
    assert meta3.success
    assert meta1.inserted_idx == [0]
    assert meta2.inserted_idx == [0]
    assert meta3.inserted_idx == [0]
    assert meta1.existing_idx == []
    assert meta2.existing_idx == []
    assert meta3.existing_idx == []


common_opt_spec = OptimizationSpecification(
    program="optprog2",
    keywords={"k": "value"},
    protocols=OptimizationProtocols(trajectory="none"),
    qc_specification=QCSpecification(
        program="prog2",
        driver=SinglepointDriver.hessian,
        method="hf",
        basis="def2-tzvp",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
    ),
)


def test_torsiondrive_socket_add_specification_same_0(storage_socket: SQLAlchemySocket):
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    # Try inserting again
    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == []
    assert meta.existing_idx == [0]
    assert id == id2


def test_torsiondrive_socket_add_specification_same_1(storage_socket: SQLAlchemySocket):
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    # model handling defaults
    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == []
    assert meta.existing_idx == [0]
    assert id == id2


def test_torsiondrive_socket_add_specification_same_2(storage_socket: SQLAlchemySocket):
    # some changes to the opt spec
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    # model handling defaults
    spec2 = TorsiondriveSpecification(
        program="torSIOndrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optpROg2",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(trajectory="none"),
            qc_specification=QCSpecification(
                program="PROG2",
                driver=SinglepointDriver.gradient,
                method="hf",
                basis="def2-tZVp",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
            ),
        ),
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == []
    assert meta.existing_idx == [0]
    assert id == id2


def test_torsiondrive_socket_add_specification_diff_1(storage_socket: SQLAlchemySocket):
    #  changing energy upper limit
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            energy_upper_limit=0.06,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id != id2


def test_torsiondrive_socket_add_specification_diff_2(storage_socket: SQLAlchemySocket):
    #  ordering of dihedrals
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = TorsiondriveSpecification(
        program="torsionndrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 13, 15)],
            grid_spacing=[15],
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id != id2


def test_torsiondrive_socket_add_specification_diff_3(storage_socket: SQLAlchemySocket):
    #  grid spacing
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = TorsiondriveSpecification(
        program="torsionndrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[5],
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id != id2


def test_torsiondrive_socket_add_specification_diff_4(storage_socket: SQLAlchemySocket):
    #  grid spacing
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.1,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id != id2


def test_torsiondrive_socket_add_specification_diff_5(storage_socket: SQLAlchemySocket):
    #  grid spacing
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.2,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.1,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    meta, id = storage_socket.records.torsiondrive.add_specification(spec1)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id is not None

    meta, id2 = storage_socket.records.torsiondrive.add_specification(spec2)
    assert meta.success
    assert meta.inserted_idx == [0]
    assert meta.existing_idx == []
    assert id != id2



from __future__ import annotations

from typing import TYPE_CHECKING

from qcfractal.testing_helpers import DummyJobProgress
from qcportal.serverinfo import ServerStatsQueryFilters
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket


def test_serverinfo_socket_update_stats(storage_socket: SQLAlchemySocket):
    stats = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters())
    assert len(stats) == 0

    time_0 = now_at_utc()

    # Force saving the stats
    with storage_socket.session_scope() as session:
        storage_socket.serverinfo.update_server_stats(session=session, job_progress=DummyJobProgress())

    time_1 = now_at_utc()

    stats = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters())
    assert len(stats) == 1

    assert stats[0]["molecule_count"] == 0
    assert stats[0]["outputstore_count"] == 0
    assert stats[0]["record_count"] == 0

    time_0 = now_at_utc()

    # Force saving the stats again
    with storage_socket.session_scope() as session:
        storage_socket.serverinfo.update_server_stats(session=session, job_progress=DummyJobProgress())

    # Should get the latest now
    stats2 = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters())
    assert len(stats2) == 2

    # Should return newest first
    assert stats2[1] == stats[0]
    assert stats2[0]["timestamp"] > time_0
    assert stats2[1]["timestamp"] < time_0

    # one more update
    with storage_socket.session_scope() as session:
        storage_socket.serverinfo.update_server_stats(session=session, job_progress=DummyJobProgress())

    stats = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters(before=now_at_utc()))
    assert len(stats) == 3

    stats = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters(before=now_at_utc(), after=time_1))
    assert len(stats) == 2

    stats = storage_socket.serverinfo.query_server_stats(ServerStatsQueryFilters(before=time_1))
    assert len(stats) == 1



from __future__ import annotations

import threading
import time
import uuid
from typing import TYPE_CHECKING

from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.components.internal_jobs.socket import InternalJobSocket
from qcportal.internal_jobs import InternalJobStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


# Add in another function to the internal_jobs socket for testing
def dummmy_internal_job(self, iterations: int, session, job_progress):
    for i in range(iterations):
        time.sleep(1.0)
        job_progress.update_progress(100 * ((i + 1) / iterations))
        print("Dummy internal job counter ", i)

        if job_progress.cancelled():
            return "Internal job cancelled"

    return "Internal job finished"


setattr(InternalJobSocket, "dummy_job", dummmy_internal_job)


def test_internal_jobs_socket_add_unique(storage_socket: SQLAlchemySocket):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=True
    )

    id_2 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=True
    )

    assert id_1 == id_2


def test_internal_jobs_socket_add_non_unique(storage_socket: SQLAlchemySocket):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    id_2 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    id_3 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    assert len({id_1, id_2, id_3}) == 3


def test_internal_jobs_socket_run(storage_socket: SQLAlchemySocket, session: Session):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    time_0 = now_at_utc()
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(6)
    time_1 = now_at_utc()

    try:
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.running
        assert job_1.progress > 10
        assert time_0 < job_1.last_updated < time_1

        time.sleep(8)
        time_2 = now_at_utc()

        session.expire(job_1)
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.complete
        assert job_1.progress == 100
        assert job_1.result == "Internal job finished"
        assert time_1 < job_1.ended_date < time_2
        assert time_1 < job_1.last_updated < time_2

    finally:
        end_event.set()
        th.join()


def test_internal_jobs_socket_recover(storage_socket: SQLAlchemySocket, session: Session):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(3)

    # Cancel/close the job runner
    end_event.set()
    th.join(20)
    assert not th.is_alive()

    job_1 = session.get(InternalJobORM, id_1)
    assert job_1.status == InternalJobStatusEnum.running
    assert job_1.progress > 10
    old_uuid = job_1.runner_uuid

    # Change uuid
    storage_socket.internal_jobs._uuid = str(uuid.uuid4())

    # Job is now running but orphaned. Should be picked up next time
    time.sleep(15)
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(15)

    try:
        session.expire(job_1)
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.complete
        assert job_1.runner_uuid != old_uuid
        assert job_1.progress == 100
        assert job_1.result == "Internal job finished"
    finally:
        end_event.set()
        th.join()



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    select,
    UniqueConstraint,
    Index,
    CheckConstraint,
    Column,
    Integer,
    ForeignKey,
    String,
    JSON,
    event,
    DDL,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, column_property

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.optimization.record_db_models import (
    OptimizationSpecificationORM,
    OptimizationRecordORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class GridoptimizationOptimizationORM(BaseORM):
    """
    Table for storing gridoptimization to optimization relationships
    """

    __tablename__ = "gridoptimization_optimization"

    gridoptimization_id = Column(
        Integer, ForeignKey("gridoptimization_record.id", ondelete="cascade"), primary_key=True
    )
    optimization_id = Column(Integer, ForeignKey("optimization_record.id"), nullable=False)
    key = Column(String, nullable=False, primary_key=True)

    energy = column_property(
        select(OptimizationRecordORM.energies[-1]).where(OptimizationRecordORM.id == optimization_id).scalar_subquery()
    )

    optimization_record = relationship(OptimizationRecordORM)

    __table_args__ = (Index("ix_gridoptimization_optimization_id", "optimization_id"),)

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "gridoptimization_id")
        return BaseORM.model_dict(self, exclude)


class GridoptimizationSpecificationORM(BaseORM):
    """
    Table for storing gridoptimization specifications
    """

    __tablename__ = "gridoptimization_specification"

    id = Column(Integer, primary_key=True)

    program = Column(String(100), nullable=False)

    optimization_specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=False)
    optimization_specification = relationship(OptimizationSpecificationORM, lazy="joined")

    keywords = Column(JSONB, nullable=False)
    keywords_hash = Column(String, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "program",
            "optimization_specification_id",
            "keywords_hash",
            name="ux_gridoptimization_specification_keys",
        ),
        Index("ix_gridoptimization_specification_program", "program"),
        Index("ix_gridoptimization_specification_optimization_specification_id", "optimization_specification_id"),
        # Enforce lowercase on some fields
        # This does not actually change the text to lowercase, but will fail to insert anything not lowercase
        # WARNING - these are not autodetected by alembic
        CheckConstraint("program = LOWER(program)", name="ck_gridoptimization_specification_program_lower"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "id", "keywords_hash", "optimization_specification_id")
        return BaseORM.model_dict(self, exclude)

    @property
    def short_description(self) -> str:
        return f"{self.program}~{self.optimization_specification.short_description}"


class GridoptimizationRecordORM(BaseRecordORM):
    """
    Table for storing gridoptimization calculations
    """

    __tablename__ = "gridoptimization_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(GridoptimizationSpecificationORM.id), nullable=False)
    specification = relationship(GridoptimizationSpecificationORM, lazy="selectin")

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    initial_molecule = relationship(MoleculeORM, foreign_keys=initial_molecule_id)

    starting_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=True)
    starting_molecule = relationship(MoleculeORM, foreign_keys=starting_molecule_id)

    starting_grid = Column(JSON)  # tuple

    optimizations = relationship(GridoptimizationOptimizationORM, cascade="all, delete-orphan", passive_deletes=True)

    __mapper_args__ = {
        "polymorphic_identity": "gridoptimization",
    }

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "specification_id")
        return BaseRecordORM.model_dict(self, exclude)

    @property
    def short_description(self) -> str:
        return f'{self.initial_molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_gridoptimization_record_delete_base_tr
    AFTER DELETE ON gridoptimization_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(
    GridoptimizationRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql"))
)



import os

from sqlalchemy import select, create_engine, Column, String, ForeignKey, LargeBinary
from sqlalchemy.orm import selectinload, sessionmaker, declarative_base

from qcfractal.db_socket.socket import SQLAlchemySocket
from qcportal.compression import compress, CompressionEnum
from qcportal.serialization import serialize

ViewBaseORM = declarative_base()


class DatasetViewEntry(ViewBaseORM):
    __tablename__ = "dataset_entry"
    name = Column(String, primary_key=True)
    data = Column(LargeBinary, nullable=False)


class DatasetViewSpecification(ViewBaseORM):
    __tablename__ = "dataset_specification"
    name = Column(String, primary_key=True)
    data = Column(LargeBinary, nullable=False)


class DatasetViewRecord(ViewBaseORM):
    __tablename__ = "dataset_record"
    entry_name = Column(String, ForeignKey(DatasetViewEntry.name), primary_key=True)
    specification_name = Column(String, ForeignKey(DatasetViewSpecification.name), primary_key=True)
    data = Column(LargeBinary, nullable=False)


class DatasetViewMetadata(ViewBaseORM):
    __tablename__ = "dataset_metadata"

    key = Column(String, ForeignKey(DatasetViewEntry.name), primary_key=True)
    value = Column(LargeBinary, nullable=False)


def _serialize_orm(orm, exclude=None):
    s_data = serialize(orm.model_dict(exclude=exclude), "application/msgpack")
    c_data, _, _ = compress(s_data, CompressionEnum.zstd, 7)
    return c_data


def create_dataset_view(dataset_id: int, socket: SQLAlchemySocket, view_file_path: str):
    if os.path.exists(view_file_path):
        raise RuntimeError(f"File {view_file_path} exists - will not overwrite")

    if os.path.isdir(view_file_path):
        raise RuntimeError(f"{view_file_path} is a directory")

    uri = "sqlite:///" + view_file_path
    engine = create_engine(uri)
    ViewSession = sessionmaker(bind=engine)

    ViewBaseORM.metadata.create_all(engine)

    view_session = ViewSession()

    with socket.session_scope(True) as fractal_session:
        ds_type = socket.datasets.lookup_type(dataset_id)
        ds_socket = socket.datasets.get_socket(ds_type)

        dataset_orm = ds_socket.dataset_orm
        entry_orm = ds_socket.entry_orm
        specification_orm = ds_socket.specification_orm
        record_item_orm = ds_socket.record_item_orm

        stmt = select(dataset_orm).where(dataset_orm.id == dataset_id)
        stmt = stmt.options(selectinload("*"))
        ds_orm = fractal_session.execute(stmt).scalar_one()

        # Metadata
        metadata_bytes = _serialize_orm(ds_orm)
        metadata_orm = DatasetViewMetadata(key="raw_data", value=metadata_bytes)
        view_session.add(metadata_orm)
        view_session.commit()

        # Entries
        stmt = select(entry_orm)
        stmt = stmt.options(selectinload("*"))
        stmt = stmt.where(entry_orm.dataset_id == dataset_id)
        entries = fractal_session.execute(stmt).scalars().all()

        for entry in entries:
            entry_bytes = _serialize_orm(entry)
            entry_orm = DatasetViewEntry(name=entry.name, data=entry_bytes)
            view_session.add(entry_orm)

        view_session.commit()

        # Specifications
        stmt = select(specification_orm)
        stmt = stmt.options(selectinload("*"))
        stmt = stmt.where(specification_orm.dataset_id == dataset_id)
        specs = fractal_session.execute(stmt).scalars().all()

        for spec in specs:
            spec_bytes = _serialize_orm(spec)
            specification_orm = DatasetViewSpecification(name=spec.name, data=spec_bytes)
            view_session.add(specification_orm)

        view_session.commit()

        base_stmt = select(record_item_orm).where(record_item_orm.dataset_id == dataset_id)
        base_stmt = base_stmt.options(selectinload("*"))
        base_stmt = base_stmt.order_by(record_item_orm.record_id.asc())

        skip = 0
        while True:
            stmt = base_stmt.offset(skip).limit(10)
            batch = fractal_session.execute(stmt).scalars()

            count = 0
            for item_orm in batch:
                item_bytes = _serialize_orm(item_orm)

                view_record_orm = DatasetViewRecord(
                    entry_name=item_orm.entry_name,
                    specification_name=item_orm.specification_name,
                    data=item_bytes,
                )

                view_session.add(view_record_orm)
                count += 1

            view_session.commit()

            if count == 0:
                break

            skip += count



from __future__ import annotations

from typing import Optional, Iterable, Dict, Any

from sqlalchemy import Column, Integer, TIMESTAMP, String, Float, BigInteger, JSON, Index, CHAR, ForeignKey
from sqlalchemy.dialects.postgresql import INET
from sqlalchemy.orm import relationship

from qcfractal.components.auth.db_models import UserORM, UserIDMapSubquery
from qcfractal.db_socket import BaseORM
from qcportal.utils import now_at_utc


class ServerStatsMetadataORM(BaseORM):
    __tablename__ = "server_stats_metadata"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    date_value = Column(TIMESTAMP(timezone=True), nullable=False)


class AccessLogORM(BaseORM):
    """
    Table for storing a log of accesses/requests
    """

    __tablename__ = "access_log"

    id = Column(Integer, primary_key=True)
    timestamp = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    method = Column(String, nullable=False)
    module = Column(String, nullable=True)
    full_uri = Column(String, nullable=True)

    request_duration = Column(Float, nullable=False)
    request_bytes = Column(BigInteger, nullable=False)
    response_bytes = Column(BigInteger, nullable=False)

    user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="AccessLogORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
    )

    # user info
    ip_address = Column(INET)
    user_agent = Column(String)

    # extra computed geo data
    country_code = Column(CHAR(2))
    subdivision = Column(String)
    city = Column(String)
    ip_lat = Column(Float)
    ip_long = Column(Float)

    __table_args__ = (
        Index("ix_access_log_timestamp", "timestamp", postgresql_using="brin"),
        Index("ix_access_log_module", "module"),
        Index("ix_access_log_user_id", "user_id"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # strip user/group ids
        exclude = self.append_exclude(exclude, "user_id")

        d = BaseORM.model_dict(self, exclude)
        d["user"] = self.user.username if self.user is not None else None
        return d


class InternalErrorLogORM(BaseORM):
    """
    Table for storing internal errors

    Internal errors are usually not reported to the user, and are stored here instead
    for retrieval by an admin
    """

    __tablename__ = "internal_error_log"

    id = Column(Integer, primary_key=True)
    error_date = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    qcfractal_version = Column(String, nullable=False)
    error_text = Column(String)
    user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="InternalErrorLogORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
    )

    request_path = Column(String)
    request_headers = Column(String)
    request_body = Column(String)

    __table_args__ = (
        Index("ix_internal_error_log_error_date", "error_date", postgresql_using="brin"),
        Index("ix_internal_error_log_user_id", "user_id"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # strip user/group ids
        exclude = self.append_exclude(exclude, "user_id")

        d = BaseORM.model_dict(self, exclude)
        d["user"] = self.user.username if self.user is not None else None
        return d


class ServerStatsLogORM(BaseORM):
    """
    Table for storing server statistics

    Server statistics (storage size, row count, etc) are periodically captured and
    stored in this table
    """

    __tablename__ = "server_stats_log"

    id = Column(Integer, primary_key=True)
    timestamp = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)

    # Raw counts
    collection_count = Column(Integer)
    molecule_count = Column(BigInteger)
    record_count = Column(BigInteger)
    outputstore_count = Column(BigInteger)
    access_count = Column(BigInteger)
    error_count = Column(BigInteger)

    # Task & service queue status
    task_queue_status = Column(JSON)
    service_queue_status = Column(JSON)

    # Database
    db_total_size = Column(BigInteger)
    db_table_size = Column(BigInteger)
    db_index_size = Column(BigInteger)
    db_table_information = Column(JSON)

    __table_args__ = (Index("ix_server_stats_log_timestamp", "timestamp", postgresql_using="brin"),)


class MessageOfTheDayORM(BaseORM):
    """
    Table for storing the Message-of-the-Day
    """

    __tablename__ = "motd"

    id = Column(Integer, primary_key=True)
    motd = Column(String, nullable=False)



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional
from urllib.parse import urlparse

from flask import request, g, current_app
from flask_jwt_extended import (
    verify_jwt_in_request,
    get_jwt,
    get_jwt_identity,
    create_access_token,
    create_refresh_token,
)
from werkzeug.exceptions import BadRequest, Forbidden

from qcfractal.flask_app import storage_socket
from qcportal.auth import UserInfo, RoleInfo
from qcportal.exceptions import AuthorizationFailure, AuthenticationFailure

if TYPE_CHECKING:
    from typing import Set

_all_endpoints: Set[str] = set()


def get_all_endpoints() -> Set[str]:
    """
    Get a list of all endpoints on the server

    These endpoints are the first three parts of the resource (ie,
    /api/v1/molecules, not /api/v1/molecules/bulkGet)
    """
    global _all_endpoints

    if not _all_endpoints:
        for url in current_app.url_map.iter_rules():
            endpoint = get_url_major_component(url.rule)
            # Don't add "static"
            if not endpoint.startswith("/static"):
                _all_endpoints.add(endpoint)

    return _all_endpoints


def get_url_major_component(url: str):
    """
    Obtains the major parts of a URL's components

    For example, /api/v1/molecule/a/b/c -> /api/v1/molecule
    """

    components = urlparse(url).path.split("/")
    resource = "/".join(components[:4])

    # Force leading slash, but only one
    return "/" + resource.lstrip("/")


def assert_role_permissions(requested_action: str):
    """
    Check for access to the URL given permissions in the JWT token in the request headers

    1. If no security (enable_security is False), always allow
    2. If security is enabled, and if read allowed (allow_unauthenticated_read=True), use the default read permissions.
       Otherwise, check against the logged-in user permissions from the headers' JWT token
    """

    # Check for the JWT in the header
    # don't raise exception if no JWT is found
    verify_jwt_in_request(optional=True)

    try:
        # TODO - some of these may not be None in the future
        claims = get_jwt()
        user_id = get_jwt_identity()  # may be None
        username = claims.get("username", None)
        policies = claims.get("permissions", {})
        role = claims.get("role", None)
        groups = claims.get("groups", None)

        subject = {"user_id": user_id, "username": username}

        # Pull the first part of the URL (ie, /api/v1/molecule/a/b/c -> /api/v1/molecule)
        resource = {"type": get_url_major_component(request.url)}

        storage_socket.auth.assert_authorized(
            resource=resource, action=requested_action, subject=subject, context={}, policies=policies
        )

        # Store the user in the global app/request context
        g.user_id = user_id
        g.username = username
        g.role = role
        g.groups = groups

    except AuthorizationFailure as e:
        raise Forbidden(str(e))
    except Exception as e:
        current_app.logger.info("Error in evaluating JWT permissions: \n" + str(e))
        raise BadRequest("Error in evaluating JWT permissions")


def access_token_from_user(user_info: UserInfo, role_info: RoleInfo):
    """
    Creates a JWT access token from user/role information
    """
    return create_access_token(
        identity=user_info.id,
        additional_claims={
            "username": user_info.username,
            "role": user_info.role,
            "groups": user_info.groups,
            "permissions": role_info.permissions.dict(),
        },
    )


def login_and_get_jwt(get_refresh_token: bool) -> Tuple[str, Optional[str]]:
    """
    Handle a login from flask

    This function authenticates the username/password sent to flask, and returns the JWT tokens.
    It handles the username/password being stored in json as well as form data.

    If get_refresh_token is True, then a refresh token is also return. Otherwise, None is returned
    for the refresh token.

    Returns
    -------
    :
        The access token and optionally the refresh token
    """
    try:
        if request.is_json:
            username = request.json["username"]
            password = request.json["password"]
        else:
            username = request.form["username"]
            password = request.form["password"]
    except Exception:
        current_app.logger.info("Invalid/malformed login request")
        raise AuthenticationFailure("Invalid/malformed login request")

    if username is None:
        current_app.logger.info("No username provided for login")
        raise AuthenticationFailure("No username provided for login")
    if password is None:
        current_app.logger.info(f"No password provided for login of user {username}")
        raise AuthenticationFailure("No password provided for login")

    try:
        user_info, role_info = storage_socket.auth.authenticate(username, password)

        # Used for logging (in the after_request_func)
        g.user_id = user_info.id

    except AuthenticationFailure as e:
        current_app.logger.info(f"Authentication failed for user {username}: {str(e)}")
        raise

    access_token = access_token_from_user(user_info, role_info)

    if get_refresh_token:
        refresh_token = create_refresh_token(identity=user_info.id)
    else:
        refresh_token = None

    current_app.logger.info(f"Successful login for user {username}")
    return access_token, refresh_token



from qcfractal.flask_app.dashboard_v1.blueprint import dashboard_v1


@dashboard_v1.route("/home", methods=["GET"])
def dashboard_home():
    return "<html><h1>Hi!</h1></html>", 200

    # return render_template("index.html",
    #                       posts={
    #                           'error_log': [],
    #                           'server_information': [],
    #                           'users_set': [],
    #                       })



from typing import Dict

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.api_v1.helpers import wrap_route
from qcportal.base_models import ProjURLParameters
from qcportal.dataset_models import (
    DatasetAddBody,
    DatasetQueryModel,
    DatasetFetchRecordsBody,
    DatasetFetchEntryBody,
    DatasetFetchSpecificationBody,
    DatasetSubmitBody,
    DatasetDeleteStrBody,
    DatasetRecordModifyBody,
    DatasetRemoveRecordsBody,
    DatasetRecordRevertBody,
    DatasetModifyMetadata,
    DatasetQueryRecords,
    DatasetDeleteParams,
)
from qcportal.exceptions import LimitExceededError


@api_v1.route("/datasets", methods=["GET"])
@wrap_route("READ")
def list_dataset_v1():
    return storage_socket.datasets.list()


@api_v1.route("/datasets/<int:dataset_id>", methods=["GET"])
@wrap_route("READ")
def get_general_dataset_v1(dataset_id: int, url_params: ProjURLParameters):
    with storage_socket.session_scope(True) as session:
        ds_type = storage_socket.datasets.lookup_type(dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)
        return ds_socket.get(dataset_id, url_params.include, url_params.exclude, session=session)


@api_v1.route("/datasets/query", methods=["POST"])
@wrap_route("READ")
def query_general_dataset_v1(body_data: DatasetQueryModel):
    with storage_socket.session_scope(True) as session:
        dataset_id = storage_socket.datasets.lookup_id(body_data.dataset_type, body_data.dataset_name, session=session)
        ds_type = storage_socket.datasets.lookup_type(dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)
        return ds_socket.get(dataset_id, body_data.include, body_data.exclude, session=session)


@api_v1.route("/datasets/queryrecords", methods=["POST"])
@wrap_route("READ")
def query_dataset_records_v1(body_data: DatasetQueryRecords):
    return storage_socket.datasets.query_dataset_records(
        record_id=body_data.record_id, dataset_type=body_data.dataset_type
    )


@api_v1.route("/datasets/<int:dataset_id>", methods=["DELETE"])
@wrap_route("DELETE")
def delete_dataset_v1(dataset_id: int, url_params: DatasetDeleteParams):
    with storage_socket.session_scope(True) as session:
        ds_type = storage_socket.datasets.lookup_type(dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)
        return ds_socket.delete_dataset(dataset_id, url_params.delete_records)


#################################################################
# COMMON FUNCTIONS
# These functions are common to all datasets
# Note that the inputs are all the same, but the returned dicts
# are different
#################################################################


########################
# Adding a dataset
########################
@api_v1.route("/datasets/<string:dataset_type>", methods=["POST"])
@wrap_route("WRITE")
def add_dataset_v1(dataset_type: str, body_data: DatasetAddBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.add(
        name=body_data.name,
        description=body_data.description,
        tagline=body_data.tagline,
        tags=body_data.tags,
        group=body_data.group,
        provenance=body_data.provenance,
        visibility=body_data.visibility,
        default_tag=body_data.default_tag,
        default_priority=body_data.default_priority,
        metadata=body_data.metadata,
        owner_user=g.username,
        owner_group=body_data.owner_group,
        existing_ok=body_data.existing_ok,
    )


#########################
# Getting info
#########################
@api_v1.route("/datasets/<int:dataset_id>", methods=["GET"])
@wrap_route("READ")
def get_dataset_general_v1(dataset_id: int, url_params: ProjURLParameters):
    return storage_socket.datasets.get(
        dataset_id,
        url_params.include,
        url_params.exclude,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>", methods=["GET"])
@wrap_route("READ")
def get_dataset_v1(dataset_type: str, dataset_id: int, url_params: ProjURLParameters):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get(
        dataset_id,
        url_params.include,
        url_params.exclude,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/status", methods=["GET"])
@wrap_route("READ")
def get_dataset_status_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.status(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/detailed_status", methods=["GET"])
@wrap_route("READ")
def get_dataset_detailed_status_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.detailed_status(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/record_count", methods=["GET"])
@wrap_route("READ")
def get_dataset_record_count_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get_record_count(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/computed_properties", methods=["GET"])
@wrap_route("READ")
def get_dataset_computed_properties_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get_computed_properties(dataset_id)


#########################
# Modifying metadata
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>", methods=["PATCH"])
@wrap_route("WRITE")
def modify_dataset_metadata_v1(dataset_type: str, dataset_id: int, body_data: DatasetModifyMetadata):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.update_metadata(dataset_id, new_metadata=body_data)


#########################
# Computation submission
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/submit", methods=["POST"])
@wrap_route("WRITE")
def submit_dataset_v1(dataset_type: str, dataset_id: int, body_data: DatasetSubmitBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.submit(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        tag=body_data.tag,
        priority=body_data.priority,
        owner_user=g.username,
        owner_group=body_data.owner_group,
        find_existing=body_data.find_existing,
    )


###################
# Specifications
###################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specification_names", methods=["GET"])
@wrap_route("READ")
def fetch_dataset_specification_names_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specification_names(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications", methods=["GET"])
@wrap_route("READ")
def fetch_all_dataset_specifications_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specifications(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications/bulkFetch", methods=["POST"])
@wrap_route("READ")
def fetch_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchSpecificationBody):
    # use the entry limit I guess?
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_dataset_entries

    if len(body_data.names) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.names)} dataset specifications - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specifications(
        dataset_id,
        specification_names=body_data.names,
        missing_ok=body_data.missing_ok,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications/bulkDelete", methods=["POST"])
@wrap_route("DELETE")
def delete_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: DatasetDeleteStrBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.delete_specifications(dataset_id, body_data.names, body_data.delete_records)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications", methods=["PATCH"])
@wrap_route("WRITE")
def rename_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: Dict[str, str]):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.rename_specifications(dataset_id, body_data)


###################
# Entries
###################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entry_names", methods=["GET"])
@wrap_route("READ")
def fetch_dataset_entry_names_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_entry_names(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries/bulkDelete", methods=["POST"])
@wrap_route("DELETE")
def delete_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetDeleteStrBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.delete_entries(dataset_id, body_data.names, body_data.delete_records)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries/bulkFetch", methods=["POST"])
@wrap_route("READ")
def fetch_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchEntryBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_dataset_entries

    if len(body_data.names) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.names)} dataset entries - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_entries(
        dataset_id,
        entry_names=body_data.names,
        missing_ok=body_data.missing_ok,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries", methods=["PATCH"])
@wrap_route("WRITE")
def rename_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: Dict[str, str]):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.rename_entries(dataset_id, body_data)


#########################
# Records
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/bulkFetch", methods=["POST"])
@wrap_route("READ")
def fetch_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchRecordsBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records

    n_requested = len(body_data.entry_names) * len(body_data.specification_names)
    if n_requested > limit:
        raise LimitExceededError(f"Cannot get {n_requested} dataset records - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)

    return ds_socket.fetch_records(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        status=body_data.status,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/bulkDelete", methods=["POST"])
@wrap_route("DELETE")
def remove_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRemoveRecordsBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.remove_records(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        delete_records=body_data.delete_records,
    )


#########################
# Record modification
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records", methods=["PATCH"])
@wrap_route("WRITE")
def modify_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRecordModifyBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.modify_records(
        dataset_id,
        g.username,
        body_data.entry_names,
        body_data.specification_names,
        body_data.status,
        body_data.priority,
        body_data.tag,
        body_data.comment,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/revert", methods=["POST"])
@wrap_route("WRITE")
def revert_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRecordRevertBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.revert_records(
        dataset_id,
        body_data.revert_status,
        body_data.entry_names,
        body_data.specification_names,
    )


###################
# Contributed Values
###################
@api_v1.route("/datasets/<int:dataset_id>/contributed_values", methods=["GET"])
@wrap_route("READ")
def fetch_dataset_contributed_values_v1(dataset_id: int):
    return storage_socket.datasets.get_contributed_values(dataset_id)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import select

from qcportal.auth import RoleInfo, is_valid_rolename
from qcportal.exceptions import UserManagementError
from .db_models import RoleORM

if TYPE_CHECKING:
    from typing import Dict, List, Any, Optional
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket

"""
Default roles are:
    * admin   (read/write to everything)
    * read    (read only, all resources except user, role, manager, access, error)
    * monitor (read only, all resources except user, role)
    * compute (for compute workers, read/write on queue_manager)
    * submit  (read on all except user, role, manager, access, error
               write on task_queue, service_queue, molecule, keyword, collection)
"""

default_roles: Dict[str, Any] = {
    "admin": {
        "Statement": [
            {"Effect": "Allow", "Action": "*", "Resource": "*"},
        ]
    },
    "read": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": ["/api/v1/users", "/api/v1/me"]},
            {
                "Effect": "Deny",
                "Action": "*",
                "Resource": [
                    "/api/v1/roles",
                    "/api/v1/managers",
                    "/api/v1/server_errors",
                    "/api/v1/access_logs",
                    "/api/v1/tasks",
                    "/api/v1/internal_jobs",
                ],
            },
        ]
    },
    "monitor": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": "/api/v1/users"},
            {"Effect": "Deny", "Action": "*", "Resource": ["/api/v1/roles"]},
        ]
    },
    "compute": {
        "Statement": [
            {"Effect": "Allow", "Action": ["READ"], "Resource": "/api/v1/information"},
            {"Effect": "Allow", "Action": ["READ", "WRITE"], "Resource": "/api/v1/users"},
            {"Effect": "Allow", "Action": "*", "Resource": ["/compute/v1/managers", "/compute/v1/tasks"]},
        ]
    },
    "submit": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": "/api/v1/users"},
            {
                "Effect": "Deny",
                "Action": "*",
                "Resource": [
                    "/api/v1/roles",
                    "/api/v1/managers",
                    "/api/v1/server_stats",
                    "/api/v1/server_errors",
                    "/api/v1/access_logs",
                    "/api/v1/tasks",
                    "/api/v1/internal_jobs",
                ],
            },
            {
                "Effect": "Allow",
                "Action": "*",
                "Resource": ["/api/v1/records", "/api/v1/molecules", "/api/v1/keywords", "/api/v1/datasets"],
            },
        ]
    },
}


class RoleSocket:
    """
    Socket for managing user roles
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

    def _get_internal(self, session: Session, rolename: str) -> RoleORM:
        """
        Returns a RoleORM for the given rolename

        If the user is not found, an exception is raised. The ORM is attached to the given session

        Parameters
        ----------
        session
            SQLAlchemy session to use for querying

        Returns
        -------
        :
            ORM of the specified role
        """

        is_valid_rolename(rolename)

        stmt = select(RoleORM).where(RoleORM.rolename == rolename)
        orm = session.execute(stmt).scalar_one_or_none()
        if orm is None:
            raise UserManagementError(f"Role {rolename} does not exist")

        return orm

    def list(self, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get information about all roles

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """
        with self.root_socket.optional_session(session, True) as session:
            stmt = select(RoleORM).order_by(RoleORM.id.asc())
            roles = session.execute(stmt).scalars().all()
            return [r.model_dict() for r in roles]

    def get(self, rolename: str, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Get information about a particular role

        Parameters
        ----------
        rolename
            Name of the role
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """
        with self.root_socket.optional_session(session, True) as session:
            role = self._get_internal(session, rolename)
            return role.model_dict()

    def add(self, role_info: RoleInfo, *, session: Optional[Session] = None) -> None:
        """
        Adds a new role.

        An exception is raised if the role already exists.

        Parameters
        ----------
        role_info
            Data about the new role
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        is_valid_rolename(role_info.rolename)

        try:
            with self.root_socket.optional_session(session) as session:
                role = RoleORM(rolename=role_info.rolename, permissions=role_info.permissions.dict())
                session.add(role)
        except IntegrityError:
            raise UserManagementError(f"Role {role_info.rolename} already exists")

        self._logger.info(f"Role {role_info.rolename} added")

    def modify(self, role_info: RoleInfo, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Update role's permissions.

        Parameters
        ----------
        role_info
            Data about the role (new permissions)
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # Cannot change admin role
        if role_info.rolename == "admin":
            raise UserManagementError("Cannot modify the admin role")

        with self.root_socket.optional_session(session) as session:
            role = self._get_internal(session, role_info.rolename)
            role.permissions = role_info.permissions.dict()
            session.commit()

            self._logger.info(f"Role {role_info.rolename} modified")
            return self.get(role_info.rolename, session=session)

    def delete(self, rolename: str, *, session: Optional[Session] = None) -> None:
        """
        Delete role.

        This will raise an exception if the role does not exist or is being referenced from somewhere else.

        Parameters
        ----------
        rolename
            The role name to delete
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # Cannot delete admin role
        if rolename == "admin":
            raise UserManagementError("Cannot delete the admin role")

        try:
            with self.root_socket.optional_session(session) as session:
                role = self._get_internal(session, rolename)
                session.delete(role)
        except IntegrityError:
            raise UserManagementError("Role could not be deleted. Likely it is being referenced somewhere")

        self._logger.info(f"Role {rolename} deleted")

    def reset_defaults(self, *, session: Optional[Session] = None) -> None:
        """
        Reset the permissions of the default roles back to their original values

        If a role does not exist, it will be created. Manually-created roles will be left alone.

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            for rolename, permissions in default_roles.items():
                stmt = select(RoleORM).where(RoleORM.rolename == rolename)
                role_data = session.execute(stmt).scalar_one_or_none()
                if role_data is None:
                    role_data = RoleORM(rolename=rolename, permissions=permissions)
                    session.add(role_data)
                else:
                    role_data.permissions = permissions

        self._logger.warning(f"Reset all roles to defaults")



from __future__ import annotations

import logging
import logging.handlers
from typing import TYPE_CHECKING

import gunicorn.app.base
from gunicorn.glogging import Logger as GLogger

from .flask_app import create_flask_app, storage_socket

if TYPE_CHECKING:
    import queue
    import threading
    from typing import Optional
    from ..config import FractalConfig


#####################################################
# See https://docs.gunicorn.org/en/stable/custom.html
#####################################################


def post_fork_cleanup(server, worker):
    """
    Do some cleanup after forking inside gunicorn

    We use synchronous workers, which are spawned via fork(). Howver,
    this would cause multiple processes to share the same db connections.
    We must dispose of them (from the global storage_socket object).

    https://docs.sqlalchemy.org/en/14/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork
    """

    storage_socket.post_fork_cleanup()


class FractalGunicornLogger(GLogger):
    def setup(self, cfg):
        # The base constructor messes with logging. So undo the important parts
        # of those changes. That way, all logging will be unified
        self.error_log = logging.getLogger("gunicorn.error")
        self.access_log = logging.getLogger("gunicorn.access")
        self.error_log.handlers = []
        self.access_log.handlers = []
        self.error_log.propagate = True
        self.access_log.propagate = True


class FractalGunicornApp(gunicorn.app.base.BaseApplication):
    def __init__(
        self,
        qcfractal_config: FractalConfig,
        finished_queue: Optional[queue.Queue] = None,
        started_event: Optional[threading.Event] = None,
    ):
        self.qcfractal_config = qcfractal_config
        self.application = create_flask_app(qcfractal_config, finished_queue=finished_queue)
        self.started_event = started_event

        super().__init__()

    def _server_ready(self, server):
        if self.started_event is not None:
            self.started_event.set()

    def load(self):
        return self.application

    def load_config(self):
        # This must be provided. It is called from the superclass, and we need to
        # populate self.cfg
        bind = self.qcfractal_config.api.host
        port = self.qcfractal_config.api.port

        # Use the sync worker class, however if threads > 1, then the gthread worker class will
        # be automatically used instead (according to gunicorn docs)
        # https://docs.gunicorn.org/en/stable/settings.html#threads
        config = {
            "worker_class": "sync",
            "bind": f"{bind}:{port}",
            "workers": self.qcfractal_config.api.num_workers,
            "threads": self.qcfractal_config.api.num_threads_per_worker,
            "timeout": self.qcfractal_config.api.worker_timeout,
            "loglevel": self.qcfractal_config.loglevel,
            "logger_class": FractalGunicornLogger,
            "post_fork": post_fork_cleanup,
            "when_ready": self._server_ready,
        }

        if self.qcfractal_config.api.extra_gunicorn_options:
            config.update(**self.qcfractal_config.api.extra_gunicorn_options)

        for key, value in config.items():
            self.cfg.set(key.lower(), value)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcportal.neb import NEBDataset
from qcportal.record_models import PriorityEnum
from .testing_helpers import load_test_data

if TYPE_CHECKING:
    from qcportal import PortalClient


@pytest.fixture(scope="function")
def neb_ds(submitter_client: PortalClient):
    ds = submitter_client.add_dataset(
        "neb",
        "Test dataset",
        "Test Description",
        "a Tagline",
        ["tag1", "tag2"],
        "new_group",
        {"prov_key_1": "prov_value_1"},
        True,
        "def_tag",
        PriorityEnum.low,
        {"meta_key_1": "meta_value_1"},
        "group1",
    )

    assert ds.owner_user is not None
    assert ds.owner_user == ds._client.username
    assert ds.owner_group == "group1"

    yield ds


@pytest.mark.parametrize("find_existing", [True, False])
def test_neb_dataset_client_submit(neb_ds: NEBDataset, find_existing: bool):
    input_spec_1, molecule_1, _ = load_test_data("neb_HCN_psi4_pbe_opt_diff")

    neb_ds.add_entry(name="test_molecule", initial_chain=molecule_1)
    neb_ds.add_specification("test_spec", input_spec_1, "test_specification")

    neb_ds.submit()
    assert neb_ds.status()["test_spec"]["waiting"] == 1

    neb_ds.submit()
    assert neb_ds.status()["test_spec"]["waiting"] == 1

    # Should only be one record
    record_id = 0
    for e, s, r in neb_ds.iterate_records():
        assert r.owner_user == neb_ds.owner_user
        assert r.owner_group == neb_ds.owner_group
        record_id = r.id

    # delete & re-add entry, then resubmit
    neb_ds.delete_entries(["test_molecule"])
    assert neb_ds.status() == {}

    # record still on the server?
    r = neb_ds._client.get_records(record_id)
    assert r.owner_user == neb_ds.owner_user

    # now resubmit
    neb_ds.add_entry(name="test_molecule", initial_chain=molecule_1)
    neb_ds.submit(find_existing=find_existing)
    assert neb_ds.status()["test_spec"]["waiting"] == 1

    for e, s, r in neb_ds.iterate_records():
        assert r.owner_user == neb_ds.owner_user
        assert r.owner_group == neb_ds.owner_group

        if find_existing:
            assert r.id == record_id
        else:
            assert r.id != record_id



"""
    Credit: https://github.com/Netflix-Skunkworks/policyuniverse
    Patrick Kelley <patrick@netflix.com>

"""

from .statement import Statement

# NotPrincipal
statement01 = dict(
    Effect="Allow",
    NotPrincipal={"AWS": ["arn:aws:iam::012345678910:root"]},
    Action=["rds:*"],
    Resource="*",
)

# "Principal": "value"
statement02 = dict(
    Effect="Allow",
    Principal="arn:aws:iam::012345678910:root",
    Action=["rds:*"],
    Resource="*",
)

statement03 = dict(
    Effect="Allow",
    Principal="arn:aws:iam::012345678910:root",
    Action=["rds:*"],
    Resource="*",
)

statement04 = dict(
    Effect="Allow",
    Principal=["arn:aws:iam::012345678910:root"],
    Action=["rds:*"],
    Resource="*",
)


statement06 = dict(
    Effect="Allow",
    Principal=["lambda.amazonaws.com"],
    Action=["rds:*"],
    Resource="*",
)

statement07 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement08 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement09 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement09_wildcard = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement10 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement11 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement12 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement13 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement14 = dict(Effect="Allow", Principal="*", Action=["rds:*"], Resource="*")

statement15 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement16 = dict(Effect="Deny", Principal="*", Action=["rds:*"], Resource="*")

# Bad ARN
statement17 = dict(
    Effect="Allow",
    Principal="arn:aws:iam::012345678910",
    Action=["rds:*"],
    Resource="*",
)

# ARN Like with wildcard account number
statement18 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

# StringLike with wildcard
statement19 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement20 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement21 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement22 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

statement23 = dict(
    Effect="Allow",
    Principal="*",
    Action=["kms:*"],
    Resource="*",
)

# Testing action groups
statement24 = dict(
    Effect="Allow",
    Principal="*",
    Action=["ec2:authorizesecuritygroupingress", "ec2:AuthorizeSecuritygroupEgress"],
)

# Testing action groups
statement25 = dict(
    Effect="Allow",
    Principal="*",
    Action=[
        "ec2:authorizesecuritygroupingress",
        "ec2:AuthorizeSecuritygroupEgress",
        "iam:putrolepolicy",
    ],
    Resource="*",
)

# Testing action groups
statement26 = dict(
    Effect="Allow",
    Principal="*",
    Action=["iam:putrolepolicy", "iam:listroles"],
    Resource="*",
)

# Testing ForAnyValue/ForAllValues without list
# Like statement 07, but this one shouldn't work
statement27 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

# Testing ForAnyValue/ForAllValues without list
# Like statement 10, but this one shouldn't work
statement28 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

# aws:PrincipalOrgID
statement29 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)

# aws:PrincipalOrgID Wildcard
statement30 = dict(
    Effect="Allow",
    Principal="*",
    Action=["rds:*"],
    Resource="*",
)


def test_statement_effect():
    statement = Statement(statement01)
    assert statement.effect == "Allow"


def test_statement_not_principal():
    statement = Statement(statement01)
    assert statement.uses_not_principal()


def test_statement_principals():
    statement = Statement(statement02)
    assert statement.principals == set(["arn:aws:iam::012345678910:root"])

    statement = Statement(statement03)
    assert statement.principals == set(["arn:aws:iam::012345678910:root"])

    statement = Statement(statement04)
    assert statement.principals == set(["arn:aws:iam::012345678910:root"])

    statement = Statement(statement06)
    assert statement.principals == set(["lambda.amazonaws.com"])

    statement_wo_principal = dict(statement06)
    del statement_wo_principal["Principal"]
    statement = Statement(statement_wo_principal)
    assert statement.principals == set([])



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, ForeignKey, String, ForeignKeyConstraint, UniqueConstraint, Index
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.manybody.record_db_models import ManybodySpecificationORM, ManybodyRecordORM
from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class ManybodyDatasetEntryORM(BaseORM):
    __tablename__ = "manybody_dataset_entry"

    dataset_id = Column(Integer, ForeignKey("manybody_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, primary_key=True)
    comment = Column(String)

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    additional_keywords = Column(JSONB, nullable=False)
    attributes = Column(JSONB, nullable=False)

    initial_molecule = relationship(MoleculeORM, lazy="joined")

    __table_args__ = (
        Index("ix_manybody_dataset_entry_dataset_id", "dataset_id"),
        Index("ix_manybody_dataset_entry_name", "name"),
        Index("ix_manybody_dataset_entry_initial_molecule_id", "initial_molecule_id"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "dataset_id", "initial_molecule_id")
        return BaseORM.model_dict(self, exclude)


class ManybodyDatasetSpecificationORM(BaseORM):
    __tablename__ = "manybody_dataset_specification"

    dataset_id = Column(Integer, ForeignKey("manybody_dataset.id", ondelete="cascade"), primary_key=True)
    name = Column(String, primary_key=True)
    description = Column(String, nullable=True)
    specification_id = Column(Integer, ForeignKey(ManybodySpecificationORM.id), nullable=False)

    specification = relationship(ManybodySpecificationORM)

    __table_args__ = (
        Index("ix_manybody_dataset_specification_dataset_id", "dataset_id"),
        Index("ix_manybody_dataset_specification_name", "name"),
        Index("ix_manybody_dataset_specification_specification_id", "specification_id"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "dataset_id", "specification_id")
        return BaseORM.model_dict(self, exclude)


class ManybodyDatasetRecordItemORM(BaseORM):
    __tablename__ = "manybody_dataset_record"

    dataset_id = Column(Integer, ForeignKey("manybody_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)
    specification_name = Column(String, primary_key=True)
    record_id = Column(Integer, ForeignKey(ManybodyRecordORM.id), nullable=False)

    record = relationship(ManybodyRecordORM)

    __table_args__ = (
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["manybody_dataset_entry.dataset_id", "manybody_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["manybody_dataset_specification.dataset_id", "manybody_dataset_specification.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        Index("ix_manybody_dataset_record_record_id", "record_id"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        # Remove fields not present in the model
        exclude = self.append_exclude(exclude, "dataset_id")
        return BaseORM.model_dict(self, exclude)


class ManybodyDatasetORM(BaseDatasetORM):
    __tablename__ = "manybody_dataset"

    id = Column(Integer, ForeignKey(BaseDatasetORM.id, ondelete="cascade"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "manybody",
    }



from __future__ import annotations

import logging
import traceback
from typing import TYPE_CHECKING

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcelemental.models import FailedOperation
from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import array
from sqlalchemy.orm import joinedload, aliased, Load

from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcportal.all_results import AllResultTypes
from qcportal.compression import CompressionEnum, compress
from qcportal.compression import decompress
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerStatusEnum
from qcportal.metadata_models import TaskReturnMetadata
from qcportal.record_models import RecordStatusEnum
from qcportal.utils import calculate_limit, now_at_utc
from .db_models import TaskQueueORM
from .reset_logic import should_reset

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Any


class TaskSocket:
    """
    Socket for managing tasks
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self._tasks_claim_limit = root_socket.qcf_config.api_limits.manager_tasks_claim

    def update_finished(
        self, manager_name: str, results_compressed: Dict[int, bytes], *, session: Optional[Session] = None
    ) -> TaskReturnMetadata:
        """
        Insert data from finished calculations into the database

        Parameters
        ----------
        manager_name
            The name of the manager submitting the results
        results_compressed
            Results (in QCSchema format), with the task_id as the key
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        all_task_ids = list(results_compressed.keys())

        self._logger.info("Received completed tasks from {}.".format(manager_name))
        self._logger.info("    Task ids: " + " ".join(str(x) for x in all_task_ids))

        tasks_success: List[int] = []
        tasks_failures: List[int] = []
        tasks_rejected: List[Tuple[int, str]] = []

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == manager_name)
            stmt = stmt.with_for_update(skip_locked=False)
            manager: Optional[ComputeManagerORM] = session.execute(stmt).scalar_one_or_none()

            if manager is None:
                self._logger.warning(f"Manager {manager_name} does not exist, but is trying to return tasks. Ignoring.")
                raise ComputeManagerError(f"Manager {manager_name} does not exist")

            if manager.status != ManagerStatusEnum.active:
                self._logger.warning(f"Manager {manager_name} is not active. Ignoring...")
                raise ComputeManagerError(f"Manager {manager_name} is not active")

            all_notifications: List[Tuple[int, RecordStatusEnum]] = []

            # For automatic resetting
            to_be_reset: List[int] = []

            for task_id, result_compressed in results_compressed.items():
                result_dict = decompress(result_compressed, CompressionEnum.zstd)
                result = pydantic.parse_obj_as(AllResultTypes, result_dict)

                # We load one at a time. This works well with 'with_for_update'
                # which will do row locking. This lock is released on commit or rollback
                # We are also deferring loading of the specific record tables. These will be lazy loaded
                # when they are needed in the update functions of the various record subsockets.
                # (I tried to use with_polymorphic, but it's kind of fussy and doesn't work well with innerjoin
                #  which is needed because with_for_update doesn't work with nullable left outer joins. This should
                #  be ok, even if the second select call doesn't use with_for_update, because any loading of
                #  a derived-class orm will need to access base_record, which I believe will be locked)
                stmt = select(TaskQueueORM).filter(TaskQueueORM.id == task_id)
                stmt = stmt.options(joinedload(TaskQueueORM.record, innerjoin=True))
                stmt = stmt.with_for_update(skip_locked=False)

                task_orm: Optional[TaskQueueORM] = session.execute(stmt).scalar_one_or_none()

                # Does the task exist?
                if task_orm is None:
                    self._logger.warning(f"Task id {task_id} does not exist in the task queue")
                    tasks_rejected.append((task_id, "Task does not exist in the task queue"))
                    continue

                record_orm: BaseRecordORM = task_orm.record
                record_id = record_orm.id

                # Start a nested transaction, so that we can rollback if there is an issue with
                # an individual result without releasing the lock on the manager
                nested_session = session.begin_nested()

                notify_status = None

                try:
                    #################################################################
                    # Perform some checks for consistency
                    #################################################################
                    # Is the task in the running state
                    # If so, do not attempt to modify the task queue. Just move on
                    if record_orm.status != RecordStatusEnum.running:
                        self._logger.warning(f"Record {record_id} (task {task_id}) is not in a running state")
                        tasks_rejected.append((task_id, "Task is not in a running state"))

                    # Was the manager that sent the data the one that was assigned?
                    # If so, do not attempt to modify the task queue. Just move on
                    elif record_orm.manager_name != manager_name:
                        self._logger.warning(
                            f"Record {record_id} (task {task_id}) claimed by {record_orm.manager_name}, not {manager_name}"
                        )
                        tasks_rejected.append((task_id, "Task is claimed by another manager"))

                    # Failed task returning FailedOperation
                    elif result.success is False and isinstance(result, FailedOperation):
                        self.root_socket.records.update_failed_task(session, record_orm, result, manager_name)
                        notify_status = RecordStatusEnum.error
                        tasks_failures.append(task_id)

                        # Should we automatically reset?
                        if self.root_socket.qcf_config.auto_reset.enabled:
                            if should_reset(record_orm, self.root_socket.qcf_config.auto_reset):
                                to_be_reset.append(record_id)

                    elif result.success is not True:
                        # QCEngine should always return either FailedOperation, or some result with success == True
                        msg = f"Unexpected return from manager for task {task_id}/base result {record_id}: Returned success != True, but not a FailedOperation"
                        error = {"error_type": "internal_fractal_error", "error_message": msg}
                        failed_op = FailedOperation(error=error, success=False)

                        self.root_socket.records.update_failed_task(session, record_orm, failed_op, manager_name)
                        notify_status = RecordStatusEnum.error

                        self._logger.error(msg)
                        tasks_rejected.append((task_id, "Returned success=False, but not a FailedOperation"))

                    # Manager returned a full, successful result
                    else:
                        self.root_socket.records.update_completed_task(session, record_orm, result, manager_name)

                        notify_status = RecordStatusEnum.complete
                        tasks_success.append(task_id)

                except Exception:
                    # We have no idea what was added or is pending for removal
                    # So rollback the transaction to the most recent commit
                    nested_session.rollback()

                    # Need a new nested transaction - previous one is dead
                    nested_session = session.begin_nested()

                    msg = "Internal FractalServer Error:\n" + traceback.format_exc()
                    error = {"error_type": "internal_fractal_error", "error_message": msg}
                    failed_op = FailedOperation(error=error, success=False)

                    self.root_socket.records.update_failed_task(session, record_orm, failed_op, manager_name)
                    notify_status = RecordStatusEnum.error

                    self._logger.error(msg)
                    tasks_rejected.append((task_id, "Internal server error"))

                finally:
                    # releases the SAVEPOINT, does not actually commit to the db
                    # (see SAVEPOINTS in postgres docs)
                    nested_session.commit()

                    if notify_status is not None:
                        all_notifications.append((record_id, notify_status))

            # Update the stats for the manager
            manager.successes += len(tasks_success)
            manager.failures += len(tasks_failures)
            manager.rejected += len(tasks_rejected)

            session.flush()

            # Automatically reset ones that should be reset
            if self.root_socket.qcf_config.auto_reset.enabled and to_be_reset:
                self._logger.info(f"Auto resetting {len(to_be_reset)} records")
                self.root_socket.records.reset(to_be_reset, session=session)

        # Send notifications that tasks were completed
        for record_id, notify_status in all_notifications:
            self.root_socket.notify_finished_watch(record_id, notify_status)

        self._logger.info(
            "Processed {} returned tasks ({} successful, {} failed, {} rejected).".format(
                len(results_compressed), len(tasks_success), len(tasks_failures), len(tasks_rejected)
            )
        )

        return TaskReturnMetadata(rejected_info=tasks_rejected, accepted_ids=(tasks_success + tasks_failures))

    def claim_tasks(
        self,
        manager_name: str,
        programs: Dict[str, List[str]],
        tags: List[str],
        limit: Optional[int] = None,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        """Claim/assign tasks for a manager

        Given tags and available programs/procedures on the manager, obtain
        waiting tasks to run.

        Parameters
        ----------
        manager_name
            Name of the manager requesting tasks
        tags
            Subset of tags to claim tasks for
        limit
            Maximum number of tasks that the manager can claim
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # Normally, checking limits is done in the route code. However, we really do not want a manager
        # to claim absolutely everything. So double check here
        limit = calculate_limit(self._tasks_claim_limit, limit)

        # CTE for finding the created_on from services which this record is a dependency of
        # If a record is a dependency of a service, we use either the record's created_on or the service's created_on,
        # whichever is earlier. That way a service doesn't have to wait for all other services to finish their tasks
        # before it can finish.
        br_task = aliased(BaseRecordORM)  # BaseRecord for the task
        br_svc = aliased(BaseRecordORM)  # BaseRecord for services

        least_date = func.least(br_task.created_on, func.min(br_svc.created_on)).label("created_on")
        svcdate_cte = select(br_task.id.label("record_id"), least_date)
        svcdate_cte = svcdate_cte.join(ServiceDependencyORM, ServiceDependencyORM.record_id == br_task.id)
        svcdate_cte = svcdate_cte.join(ServiceQueueORM, ServiceQueueORM.id == ServiceDependencyORM.service_id)
        svcdate_cte = svcdate_cte.join(br_svc, br_svc.id == ServiceQueueORM.record_id)
        svcdate_cte = svcdate_cte.where(br_task.status == RecordStatusEnum.waiting)
        svcdate_cte = svcdate_cte.group_by(br_task.id)
        svcdate_cte = svcdate_cte.order_by(least_date.asc())
        svcdate_cte = svcdate_cte.cte()

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == manager_name)
            stmt = stmt.with_for_update(skip_locked=False)
            manager: Optional[ComputeManagerORM] = session.execute(stmt).scalar_one_or_none()

            if manager is None:
                self._logger.warning(f"Manager {manager_name} does not exist! Will not give it tasks")
                raise ComputeManagerError("Manager does not exist!")
            elif manager.status != ManagerStatusEnum.active:
                self._logger.warning(f"Manager {manager_name} exists but is not active! Will not give it tasks")
                raise ComputeManagerError("Manager is not active!")

            manager_programs = array(programs.keys())
            found: List[Dict[str, Any]] = []

            # Remove tags that we didn't say we handled, but keep the order
            search_tags = [x for x in tags if x in manager.tags]
            for tag in search_tags:
                new_limit = limit - len(found)

                # Have we found all we needed to find
                # (should always be >= 0, but can never be too careful. If it wasn't this could be an infinite loop)
                if new_limit <= 0:
                    break

                # Find tasks/base result:
                #   1. Status is waiting
                #   2. Whose required programs I am able to match
                #   3. Whose tags I am able to compute
                # with_for_update locks the rows. skip_locked=True makes it skip already-locked rows
                # (possibly from another process)
                # Also, load the base_result object so we can update stuff there (status)
                # TODO - we only test for the presence of the available_programs in the requirements. Eventually
                #        we want to then verify the versions

                # We do a plain .join() because we are querying, and then also supplying contains_eager() so that
                # the TaskQueueORM.record gets populated
                # See https://docs-sqlalchemy.readthedocs.io/ko/latest/orm/loading_relationships.html#routing-explicit-joins-statements-into-eagerly-loaded-collections
                stmt = select(TaskQueueORM, BaseRecordORM).join(TaskQueueORM.record)

                # Only load a few columns we need of the record
                stmt = stmt.options(
                    Load(BaseRecordORM).load_only(
                        BaseRecordORM.status, BaseRecordORM.manager_name, BaseRecordORM.modified_on
                    )
                )

                stmt = stmt.join(svcdate_cte, svcdate_cte.c.record_id == BaseRecordORM.id, isouter=True)
                stmt = stmt.filter(BaseRecordORM.status == RecordStatusEnum.waiting)
                stmt = stmt.filter(manager_programs.contains(TaskQueueORM.required_programs))

                # Order by priority, then created_on (earliest first)
                # Where the created_on may be the created_on of the parent service (see CTE above)
                stmt = stmt.order_by(
                    TaskQueueORM.priority.desc(), func.least(BaseRecordORM.created_on, svcdate_cte.c.created_on).asc()
                )

                # If tag is "*", then the manager will pull anything
                if tag != "*":
                    stmt = stmt.filter(TaskQueueORM.tag == tag)

                # Skip locked rows - They may be in the process of being claimed by someone else
                stmt = stmt.limit(new_limit).with_for_update(of=[BaseRecordORM, TaskQueueORM], skip_locked=True)

                new_items = session.execute(stmt).all()

                # Update all the task records to reflect this manager claiming them
                for _, record_orm in new_items:
                    record_orm.status = RecordStatusEnum.running
                    record_orm.manager_name = manager_name
                    record_orm.modified_on = now_at_utc()

                # Store in dict form for returning, but no need to store the info from the base record
                # Also, retrieve the actual function kwargs. Eventually we may want the managers
                # to retrieve the kwargs themselves
                for task_orm, _ in new_items:
                    task_dict = task_orm.model_dict(exclude=["record"])

                    if task_orm.function is None:
                        # Generate the task on the fly
                        task_spec = self.root_socket.records.generate_task_specification(task_orm)

                        kwargs = task_spec["function_kwargs"]
                        kwargs_compressed, _, _ = compress(kwargs, CompressionEnum.zstd)

                        # Add this to the orm for any future managers claiming this task
                        task_orm.function = task_spec["function"]
                        task_orm.function_kwargs_compressed = kwargs_compressed

                        # But just use what we created when returning to this manager
                        task_dict["function"] = task_spec["function"]
                        task_dict["function_kwargs_compressed"] = kwargs_compressed

                    found.append(task_dict)

                session.flush()

            manager.claimed += len(found)

            self._logger.info(f"Manager {manager_name} has claimed {len(found)} new tasks")

        return found
